'''
Python wrapper around implementation of the Streamlined NTRU Prime cryptosystem.

To access the Python functions provided by ntruprime, import the library (for, e.g., sntrup1277):

    from ntruprime import sntrup1277

To generate a key pair:

    pk,sk = sntrup1277.keypair()

To generate a ciphertext c encapsulating a randomly generated session key k:

    c,k = sntrup1277.enc(pk)

To recover a session key from a ciphertext:

    k = sntrup1277.dec(c,sk)

As a larger example, the following test script creates a key pair, creates a ciphertext and session key, and then recovers the session key from the ciphertext:

    import ntruprime
    kem = ntruprime.sntrup1277
    pk,sk = kem.keypair()
    c,k = kem.enc(pk)
    assert k == kem.dec(c,sk)
'''

from .kem import sntrup653
from .kem import sntrup761
from .kem import sntrup857
from .kem import sntrup953
from .kem import sntrup1013
from .kem import sntrup1277



from typing import Tuple as _Tuple
import ctypes as _ct
from ._lib import _lib, _check_input


class _KEM:
    def __init__(self) -> None:
        '''
        '''
        self._c_keypair = getattr(_lib, '%s_keypair' % self._prefix)
        self._c_keypair.argtypes = [_ct.c_char_p, _ct.c_char_p]
        self._c_keypair.restype = None
        self._c_enc = getattr(_lib, '%s_enc' % self._prefix)
        self._c_enc.argtypes = [_ct.c_char_p, _ct.c_char_p, _ct.c_char_p]
        self._c_enc.restype = None
        self._c_dec = getattr(_lib, '%s_dec' % self._prefix)
        self._c_dec.argtypes = [_ct.c_char_p, _ct.c_char_p, _ct.c_char_p]
        self._c_dec.restype = None

    def keypair(self) -> _Tuple[bytes, bytes]:
        '''
        Keypair - randomly generates secret key 'sk' and corresponding public key 'pk'.
        Returns:
            pk (bytes): public key
            sk (bytes): secret key
        '''
        pk = _ct.create_string_buffer(self.PUBLICKEYBYTES)
        sk = _ct.create_string_buffer(self.SECRETKEYBYTES)
        self._c_keypair(pk, sk)
        return pk.raw, sk.raw

    def enc(self, pk: bytes) -> _Tuple[bytes, bytes]:
        '''
        Encapsulation - randomly generates a ciphertext 'c' and the corresponding session key 'k' given Alice's public key 'pk'.
        Parameters:
            pk (bytes): public key
        Returns:
            c (bytes): ciphertext
            k (bytes): session key
        '''
        _check_input(pk, self.PUBLICKEYBYTES, 'pk')
        c = _ct.create_string_buffer(self.CIPHERTEXTBYTES)
        k = _ct.create_string_buffer(self.BYTES)
        pk = _ct.create_string_buffer(pk)
        self._c_enc(c, k, pk)
        return c.raw, k.raw

    def dec(self, c: bytes, sk: bytes) -> bytes:
        '''
        Decapsulation - given Alice's secret key 'sk' computes the session key 'k' corresponding to a ciphertext 'c'.
        Parameters:
            c (bytes): ciphertext
            sk (bytes): secret key
        Returns:
            k (bytes): session key
        '''
        _check_input(c, self.CIPHERTEXTBYTES, 'c')
        _check_input(sk, self.SECRETKEYBYTES, 'sk')
        k = _ct.create_string_buffer(self.BYTES)
        c = _ct.create_string_buffer(c)
        sk = _ct.create_string_buffer(sk)
        self._c_dec(k, c, sk)
        return k.raw


class sntrup653(_KEM):
    PUBLICKEYBYTES = 994
    SECRETKEYBYTES = 1518
    CIPHERTEXTBYTES = 897
    BYTES = 32
    _prefix = "ntruprime_kem_sntrup653"


sntrup653 = sntrup653()


class sntrup761(_KEM):
    PUBLICKEYBYTES = 1158
    SECRETKEYBYTES = 1763
    CIPHERTEXTBYTES = 1039
    BYTES = 32
    _prefix = "ntruprime_kem_sntrup761"


sntrup761 = sntrup761()


class sntrup857(_KEM):
    PUBLICKEYBYTES = 1322
    SECRETKEYBYTES = 1999
    CIPHERTEXTBYTES = 1184
    BYTES = 32
    _prefix = "ntruprime_kem_sntrup857"


sntrup857 = sntrup857()


class sntrup953(_KEM):
    PUBLICKEYBYTES = 1505
    SECRETKEYBYTES = 2254
    CIPHERTEXTBYTES = 1349
    BYTES = 32
    _prefix = "ntruprime_kem_sntrup953"


sntrup953 = sntrup953()


class sntrup1013(_KEM):
    PUBLICKEYBYTES = 1623
    SECRETKEYBYTES = 2417
    CIPHERTEXTBYTES = 1455
    BYTES = 32
    _prefix = "ntruprime_kem_sntrup1013"


sntrup1013 = sntrup1013()


class sntrup1277(_KEM):
    PUBLICKEYBYTES = 2067
    SECRETKEYBYTES = 3059
    CIPHERTEXTBYTES = 1847
    BYTES = 32
    _prefix = "ntruprime_kem_sntrup1277"


sntrup1277 = sntrup1277()
